/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.model.framestore;

import java.lang.reflect.*;
import java.util.*;

import edu.stanford.smi.protege.util.*;

public class CallCachingFrameStore extends AbstractFrameStoreInvocationHandler {
    private CacheMap _cache = new CacheMap();
    private MethodCall _lookupMethodCall = new MethodCall();

    protected Object handleInvoke(Method method, Object[] args) {
        Object result;
        if (isGetEvents(method)) {
            result = invoke(method, args);
            Collection c = (Collection) result;
            if (!c.isEmpty()) {
                clearCache();
            }
        } else if (isQuery(method)) {
            result = query(method, args);
        } else {
            _cache.clear();
            result = invoke(method, args);
        }
        return result;
    }

    private boolean isGetEvents(Method m) {
        return m.getName().equals("getEvents");
    }

    private Object query(Method method, Object[] args) {
        _lookupMethodCall.set(method, args);
        Object o = _cache.get(_lookupMethodCall);
        if (o == null) {
            o = invoke(method, args);
            _cache.put(new MethodCall(method, args), o);
        }
        return o;
    }

    private void clearCache() {
        _cache.clear();
    }
    public void handleClose() {
        _cache.clear();
        _cache = null;
    }
    protected void handleReinitialize() {
        _cache.clear();
    }
}

/*
interface Queries {
    String FRAMES = "getFrames";
    String SLOTS = "getSlots";
    String FACETS = "getFacets";
    String CLASSES = "getClses";
    String INSTANCES = "getInstances";
    String REFERENCES = "getReferences";
    String DIRECT_OWN_SLOT_VALUES = "getDirectOwnSlotValues";
    String OWN_SLOT_VALUES = "getOwnSlotValues";
    String FRAME = "getFrame";
    String FRAME_NAME = "getFrameName";
    String DIRECT_ATTACHED_CLSES = "getDirectAttachedClses";
    String OWN_SLOTS = "getOwnSlots";
    String TEMPLATE_SLOTS = "getTemplateSlots";
    String DIRECT_TEMPLATE_SLOTS = "getDirectTemplateSlots";
    String DIRECT_SUPERCLASSES = "getDirectSuperclasses";
    String SUPERCLASSES = "getSuperclasses";
    String DIRECT_TYPES = "getDirectTypes";
    String TYPES = "getTypes";
    String DIRECT_INSTANCES = "getDirectInstances";
    String SUBCLASSES = "getSubclasses";
    String DIRECT_SUBCLASSES = "getDirectSubclasses";
    String TEMPLATE_SLOT_VALUES = "getTemplateSlotValues";
    String DIRECT_TEMPLATE_SLOT_VALUES = "getDirectTemplateSlotValues";
    String TEMPLATE_FACETS = "getTemplateFacets";
    String TEMPLATE_FACET_VALUES = "getTemplateFacetValues";
    String DIRECT_TEMPLATE_FACET_VALUES = "getDirectTemplateFacetValues";
    String SUBSLOTS = "getSubslots";
    String DIRECT_SUBSLOTS = "getDirectSubslots";
    String SUPERSLOTS = "getSuperslots";
    String DIRECT_SUPERSLOTS = "getDirectSuperslots";
    String MATCHING_REFERENCES = "getMatchingReferences";
    String FRAMES_WITH_OWN_SLOT_VALUE = "getFramesWithOwnSlotValue";
    String FRAMES_WITH_MATCHING_OWN_SLOT_VALUE = "getFramesWithMatchingOwnSlotValue";
    String OWN_FACETS = "getOwnFacets";
    String OWN_FACET_VALUES = "getOwnFacetValues";
    String OVERRIDDEN_TEMPLATE_SLOTS = "getOverriddenTemplateSlots";
    String DIRECTLY_OVERRIDDEN_TEMPLATE_SLOTS = "getDirectlyOverriddenTemplateSlots";
    String OVERRIDDEN_TEMPLATE_FACETS = "getOverriddenTemplateFacets";
    String DIRECTLY_OVERRIDDEN_TEMPLATE_FACETS = "getDirectlyOverriddenTemplateFacets";
}

public class CachingFrameStore extends AbstractDelegatingFrameStore implements Queries {
    private Map _cache = new HashMap();    // Map<methodName, Map<Object[], SoftReference<Object>>>
    private Map _methods = new HashMap();  // Map<methodName, Method>
    private Arguments _lookupArguments = new Arguments();

    private Object query(String methodName, Session session) throws Exception {
        return query(methodName, new Object[] {session});
    }

    private Object query(String methodName, Object o1, Session session) throws Exception {
        return query(methodName, new Object[] {o1, session});
    }

    private Object query(String methodName, Object o1, Object o2, Session session) throws Exception {
        return query(methodName, new Object[] {o1, o2, session});
    }

    private Object query(String methodName, Object o1, Object o2, Object o3, Session session) throws Exception {
        return query(methodName, new Object[] {o1, o2, o3, session});
    }

    private Object query(String methodName, Object[] args) throws Exception {
        Map methodCache = getOrCreateMethodCache(methodName);
        Object value = getCachedValue(methodCache, args);
        if (value == null) {
            Method method = getMethod(methodName);
            value = method.invoke(getDelegate(), args);
            setCachedValue(methodCache, args, value);
        }
        return value;
    }

    private Method getMethod(String name) {
        Method method = (Method) _methods.get(name);
        if (method == null) {
            boolean foundIt = false;
            Method[] methods = FrameStore.class.getMethods();
            for (int i = 0; i < methods.length; ++i) {
                if (name.equals(methods[i].getName())) {
                    if (foundIt) {
                        System.out.println("*** duplicate method name *** " + name);
                    } else {
                        method = methods[i];
                        _methods.put(name, method);
                        foundIt = true;
                    }
                }
            }
        }
        return method;
    }

    private Map getOrCreateMethodCache(String method) {
        Map methodCache = (Map) _cache.get(method);
        if (methodCache == null) {
            methodCache = new HashMap();
            _cache.put(method, methodCache);
        }
        return methodCache;
    }

    private Object getCachedValue(Map map, Object[] args) {
        _lookupArguments.set(args);
        Reference ref = (Reference) map.get(_lookupArguments);
        Object o = (ref == null) ? null : ref.get();
        if (o instanceof Collection) {
            if (o instanceof Set) {
                o = Collections.unmodifiableSet((Set) o);
            } else if (o instanceof List) {
                o = Collections.unmodifiableList((List) o);
            } else {
                throw new RuntimeException("unknown collection");
            }
        }
        return o;
    }

    private void setCachedValue(Map cache, Object [] args, Object value) {
        cache.put(new Arguments(args), new SoftReference(value));
    }

    private void setCachedValue(String method, Object arg1, Object value, Session session) {
        setCachedValue(method, new Object[] {arg1, session}, value);
    }

    private void setCachedValue(String method, Object arg1, Object arg2, Object value, Session session) {
        setCachedValue(method, new Object[] {arg1, arg2, session}, value);
    }

    private void setCachedValue(String method, Object arg1, Object arg2, Object arg3, Object value, Session session) {
        setCachedValue(method, new Object[] {arg1, arg2, arg3, session}, value);
    }

    private void setCachedValue(String method, Object[] args, Object value) {
        Map map = getOrCreateMethodCache(method);
        setCachedValue(map, args, value);
    }

    private void clearCache() {
        _cache.clear();
    }

    private void clearCache(String methodName) {
        _cache.remove(methodName);
    }

    // -------------------- queries ----------------------------------
    public Set getClses(Session session) throws Exception {
        return (Set) query(CLASSES, session);
    }

    public Set getSlots(Session session) throws Exception {
        return (Set) query(SLOTS, session);
    }

    public Set getFacets(Session session) throws Exception {
        return (Set) query(FACETS, session);
    }

    public Set getFrames(Session session) throws Exception {
        return (Set) query(FRAMES, session);
    }

    public Set getInstances(Cls cls, Session session) throws Exception {
        return (Set) query(INSTANCES, cls, session);
    }

    public Set getReferences(Frame frame, Session session) throws Exception {
        return (Set) query(REFERENCES, frame, session);
    }

    public List getDirectOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        return (List) query(DIRECT_OWN_SLOT_VALUES, frame, slot, session);
    }

    public Collection getOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        return (Collection) query(OWN_SLOT_VALUES, frame, slot, session);
    }

    public Frame getFrame(String name, Session session) throws Exception {
        return (Frame) query(FRAME, name, session);
   }

    public String getFrameName(Frame frame, Session session) throws Exception {
        return (String) query(FRAME_NAME, frame, session);
    }

    public Set getDirectAttachedClses(Slot slot, Session session) throws Exception {
        return (Set) query(DIRECT_ATTACHED_CLSES, slot, session);
    }

    public Set getOwnSlots(Frame frame, Session session) throws Exception {
        return (Set) query(OWN_SLOTS, frame, session);
    }

    public Set getTemplateSlots(Cls cls, Session session) throws Exception {
        return (Set) query(TEMPLATE_SLOTS, cls, session);
    }

    public List getDirectTemplateSlots(Cls cls, Session session) throws Exception {
        return (List) query(DIRECT_TEMPLATE_SLOTS, cls, session);
    }

    public Collection getTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        return (Collection) query(TEMPLATE_SLOT_VALUES, cls, slot, session);
    }

    public List getDirectTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        return (List) query(DIRECT_TEMPLATE_SLOT_VALUES, cls, slot, session);
    }

    public Set getTemplateFacets(Cls cls, Slot slot, Session session) throws Exception {
        return (Set) query(TEMPLATE_FACETS, cls, slot, session);
    }

    public Collection getTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        return (Collection) query(TEMPLATE_FACET_VALUES, cls, slot, facet, session);
    }

    public List getDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        return (List) query(DIRECT_TEMPLATE_FACET_VALUES, cls, slot, facet, session);
    }

    public List getDirectSuperclasses(Cls cls, Session session) throws Exception {
        return (List) query(DIRECT_SUPERCLASSES, cls, session);
    }
    public Set getSuperclasses(Cls cls, Session session) throws Exception {
        return (Set) query(SUPERCLASSES, cls, session);
    }
    public List getDirectSubclasses(Cls cls, Session session) throws Exception {
        return (List) query(DIRECT_SUBCLASSES, cls, session);
    }
    public Set getSubclasses(Cls cls, Session session) throws Exception {
        return (Set) query(SUBCLASSES, cls, session);
    }
    public List getDirectSuperslots(Slot slot, Session session) throws Exception {
        return (List) query(DIRECT_SUPERSLOTS, slot, session);
    }
    public Set getSuperslots(Slot slot, Session session) throws Exception {
        return (Set) query(SUPERSLOTS, slot, session);
    }
    public List getDirectSubslots(Slot slot, Session session) throws Exception {
        return (List) query(DIRECT_SUBSLOTS, slot, session);
    }
    public Set getSubslots(Slot slot, Session session) throws Exception {
        return (Set) query(SUBSLOTS, slot, session);
    }

    public List getDirectTypes(Instance instance, Session session) throws Exception {
        return (List) query(DIRECT_TYPES, instance, session);
    }
    public Set getTypes(Instance instance, Session session) throws Exception {
        return (Set) query(TYPES, instance, session);
    }
    public List getDirectInstances(Cls cls, Session session) throws Exception {
        return (List) query(DIRECT_INSTANCES, cls, session);
    }
    public Set getReferences(Object object, Session session) throws Exception {
        return (Set) query(REFERENCES, object, session);
    }
    public Set getMatchingReferences(String string, Session session) throws Exception {
        return (Set) query(MATCHING_REFERENCES, string, session);
    }
    public Set getFramesWithOwnSlotValue(Slot slot, Object value, Session session) throws Exception {
        return (Set) query(FRAMES_WITH_OWN_SLOT_VALUE, slot, value, session);
    }
    public Set getFramesWithMatchingOwnSlotValue(Slot slot, String value, Session session) throws Exception {
        return (Set) query(FRAMES_WITH_MATCHING_OWN_SLOT_VALUE, slot, value, session);
    }
    public Set getOwnFacets(Frame frame, Slot slot, Session session) throws Exception {
        return (Set) query(OWN_FACETS, frame, slot, session);
    }

    public Collection getOwnFacetValues(Frame frame, Slot slot, Facet facet, Session session) throws Exception {
        return (Collection) query(OWN_FACET_VALUES, frame, slot, facet, session);
    }

    public Set getOverriddenTemplateSlots(Cls cls, Session session) throws Exception {
        return (Set) query(OVERRIDDEN_TEMPLATE_SLOTS, cls, session);
    }

    public Set getDirectlyOverriddenTemplateSlots(Cls cls, Session session) throws Exception {
        return (Set) query(DIRECTLY_OVERRIDDEN_TEMPLATE_SLOTS, cls, session);
    }

    public Set getOverriddenTemplateFacets(Cls cls, Slot slot, Session session) throws Exception {
        return (Set) query(OVERRIDDEN_TEMPLATE_FACETS, cls, slot, session);
    }

    public Set getDirectlyOverriddenTemplateFacets(Cls cls, Slot slot, Session session) throws Exception {
        return (Set) query(DIRECTLY_OVERRIDDEN_TEMPLATE_FACETS, cls, slot, session);
    }

    public Set executeQuery(Query query, Session session) throws Exception {
        return getDelegate().executeQuery(query, session);
    }

    public List getEvents(Session session) throws Exception {
        return getDelegate().getEvents(session);
    }

    // modifications
    public void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values, Session session) throws Exception {
        getDelegate().setDirectOwnSlotValues(frame, slot, values, session);
        setCachedValue(DIRECT_OWN_SLOT_VALUES, frame, slot, values, session);
    }

    public void deleteFrame(Frame frame, Session session) throws Exception {
        getDelegate().deleteFrame(frame, session);
        clearCache();
    }

    public void setFrameName(Frame frame, String name, Session session) throws Exception {
        getDelegate().setFrameName(frame, name, session);
        setCachedValue(FRAME_NAME, frame, name, session);
    }

    public Cls createCls(long id, String name, Collection directTypes, Collection directSuperclasses, boolean loadDefaults, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        clearCache(DIRECT_SUBCLASSES);
        clearCache(SUBCLASSES);
        return getDelegate().createCls(id, name, directTypes, directSuperclasses, loadDefaults, session);
    }

    public Slot createSlot(long id, String name, Collection directTypes, Collection directSuperslots, boolean loadDefaults, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        clearCache(DIRECT_SUBSLOTS);
        clearCache(SUBSLOTS);
        return getDelegate().createSlot(id, name, directTypes, directSuperslots, loadDefaults, session);
    }

    public Facet createFacet(long id, String name, Collection directTypes, boolean loadDefaults, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        return getDelegate().createFacet(id, name, directTypes, loadDefaults, session);
    }

    public SimpleInstance createSimpleInstance(long id, String name, Collection directTypes, boolean loadDefaults, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        return getDelegate().createSimpleInstance(id, name, directTypes, loadDefaults, session);
    }

    public void addDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        clearCache(DIRECT_TEMPLATE_SLOTS);
        clearCache(TEMPLATE_SLOTS);
        clearCache(OWN_SLOTS);
        getDelegate().addDirectTemplateSlot(cls, slot, session);
    }

    public void removeDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        clearCache(DIRECT_TEMPLATE_SLOTS);
        clearCache(TEMPLATE_SLOTS);
        clearCache(TEMPLATE_SLOTS);
        getDelegate().removeDirectTemplateSlot(cls, slot, session);
    }

    public void moveDirectTemplateSlot(Cls cls, Slot slot, int index, Session session) throws Exception {
        clearCache(DIRECT_TEMPLATE_SLOTS);
        getDelegate().moveDirectTemplateSlot(cls, slot, index, session);
        clearCache();
    }

    public void setDirectTemplateSlotValues(Cls cls, Slot slot, Collection values, Session session) throws Exception {
        setCachedValue(DIRECT_TEMPLATE_SLOT_VALUES, cls, slot, values, session);
        clearCache(TEMPLATE_SLOT_VALUES);
        clearCache(OWN_SLOT_VALUES);
        getDelegate().setDirectTemplateSlotValues(cls, slot, values, session);
    }

    public void setDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Collection values, Session session) throws Exception {
        setCachedValue(DIRECT_TEMPLATE_FACET_VALUES, cls, slot, facet, values, session);
        clearCache(TEMPLATE_FACET_VALUES);
        getDelegate().setDirectTemplateFacetValues(cls, slot, facet, values, session);
    }

    public void addDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        clearCache(DIRECT_SUPERCLASSES);
        clearCache(DIRECT_SUBCLASSES);
        clearCache(SUPERCLASSES);
        clearCache(SUBCLASSES);
        clearCache(TEMPLATE_SLOTS);
        clearCache(OWN_SLOTS);
        getDelegate().addDirectSuperclass(cls, superclass, session);
    }

    public void removeDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        clearCache(DIRECT_SUPERCLASSES);
        clearCache(DIRECT_SUBCLASSES);
        clearCache(SUPERCLASSES);
        clearCache(SUBCLASSES);
        clearCache(TEMPLATE_SLOTS);
        clearCache(OWN_SLOTS);
        getDelegate().removeDirectSuperclass(cls, superclass, session);
        clearCache();
    }

    public void moveDirectSubclass(Cls cls, Cls subclass, int index, Session session) throws Exception {
        clearCache(DIRECT_SUBCLASSES);
        getDelegate().moveDirectSubclass(cls, subclass, index, session);
        clearCache();
    }

    public void addDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        clearCache(DIRECT_SUPERSLOTS);
        clearCache(DIRECT_SUBSLOTS);
        clearCache(SUPERSLOTS);
        clearCache(SUBSLOTS);
        clearCache(OWN_SLOT_VALUES);
        getDelegate().addDirectSuperslot(slot, superslot, session);
        clearCache();
    }

    public void removeDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        clearCache(DIRECT_SUPERSLOTS);
        clearCache(DIRECT_SUBSLOTS);
        clearCache(SUPERSLOTS);
        clearCache(SUBSLOTS);
        clearCache(OWN_SLOT_VALUES);
        getDelegate().removeDirectSuperslot(slot, superslot, session);
    }

    public void addDirectType(Instance instance, Cls type, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        clearCache(DIRECT_INSTANCES);
        clearCache(INSTANCES);
        clearCache(OWN_SLOTS);
        getDelegate().addDirectType(instance, type, session);
    }

    public void removeDirectType(Instance instance, Cls type, Session session) throws Exception {
        clearCache(DIRECT_TYPES);
        clearCache(TYPES);
        clearCache(DIRECT_INSTANCES);
        clearCache(INSTANCES);
        clearCache(OWN_SLOTS);
        getDelegate().removeDirectType(instance, type, session);
    }

    public void removeDirectTemplateFacetOverrides(Cls cls, Slot slot, Session session) throws Exception {
        clearCache(DIRECT_TEMPLATE_FACET_VALUES);
        clearCache(TEMPLATE_FACET_VALUES);
        clearCache(OWN_FACET_VALUES);
        clearCache(DIRECTLY_OVERRIDDEN_TEMPLATE_SLOTS);
        clearCache(OVERRIDDEN_TEMPLATE_SLOTS);
        getDelegate().removeDirectTemplateFacetOverrides(cls, slot, session);
    }
}
/*
public class CachingFrameStoreHandler extends DelegatingFrameStoreHandler {
    private Map _cache = new HashMap(); // Map<Method, Map<Object[], SoftReference<Object>>>

    public Object handleInvoke(Method method, Object[] args) throws Exception {
        Object o;
        if (isQuery(method)) {
            o = handleQuery(method, args);
        } else {
            o = handleModification(method, args);
        }
        return o;
    }

    private Object handleQuery(Method method, Object[] args) throws Exception {
        Map methodCache = getOrCreateMethodCache(method);
        Object value = getCachedValue(methodCache, args);
        if (value == null) {
            value = method.invoke(getDelegate(), args);
            setCachedValue(methodCache, args, value);
        }
        return value;
    }

    private Map getOrCreateMethodCache(Method method) {
        Map methodCache = (Map) _cache.get(method);
        if (methodCache == null) {
            methodCache = new HashMap();
            _cache.put(method, methodCache);
        }
        return methodCache;
    }

    private Object getCachedValue(Map map, Object[] args) {
        Reference ref = (Reference) map.get(args);
        return (ref == null) ? null : ref.get();
    }

    private void setCachedValue(Map map, Object[] args, Object value) {
        map.put(args, new SoftReference(value));
    }

    private Object handleModification(Method method, Object[] args) throws Exception {
        updateCache(method, args);
        return method.invoke(getDelegate(), args);
    }

    private void updateCache(Method method, Object[] args) {
        _cache.clear();
    }
}
*/

/*
public class CachingFrameStore extends AbstractDelegatingFrameStore {
    private Reference _clsesCache;
    private Reference _slotsCache;
    private Reference _facetsCache;
    private Map _referencesCache = new HashMap();
    private Map _directOwnSlotValuesCache = new HashMap();
    private Map _ownSlotValuesCache = new HashMap();
    private Map _frameCache = new HashMap();
    private Map _nameCache = new HashMap();
    private Map _directSuperclassesCache = new HashMap();
    private Map _directSubclassesCache = new HashMap();
    private Map _superclassesCache = new HashMap();
    private Map _subclassesCache = new HashMap();
    private Map _directSuperslotsCache = new HashMap();
    private Map _directSubslotsCache = new HashMap();
    private Map _superslotsCache = new HashMap();
    private Map _subslotsCache = new HashMap();
    private Map _directTypesCache = new HashMap();
    private Map _directInstancesCache = new HashMap();
    private Map _typesCache = new HashMap();
    private Map _instancesCache = new HashMap();
    private Map _associatedFacetCache = new HashMap();
    private Map _associatedSlotCache = new HashMap();

    private Pair _lookupPair = new Pair(this, this);
    private Triple _lookupTriple = new Triple(this, this, this);

    private Object getCachedValue(Reference ref) {
        return (ref == null) ? null : ref.get();
    }

    private Object getCachedValue(Object key, Map map) {
        Reference ref = (Reference) map.get(key);
        return getCachedValue(ref);
    }

    private Object getCachedValue(Object o1, Object o2, Map map) {
        _lookupPair.set(o1, o2);
        Reference ref = (Reference) map.get(_lookupPair);
        return getCachedValue(ref);
    }

    private void setCachedValue(Object key, Object value, Map map) {
        map.put(key, createReference(value));
    }

    private void setCachedValue(Object o1, Object o2, Object value, Map map) {
        map.put(new Pair(o1, o2), createReference(value));
    }

    private void removeCachedValue(Object o1, Object o2, Map map) {
        _lookupPair.set(o1, o2);
        map.remove(_lookupPair);
    }

    private Reference createReference(Object value) {
        return new SoftReference(value);
    }

    // -------------------- queries ----------------------------------
    public Set getClses(Session session) throws Exception {
        Set clses = (Set) getCachedValue(_clsesCache);
        if (clses == null) {
            clses = getDelegate().getClses(session);
            _clsesCache = createReference(clses);
        }
        return clses;
    }

    public Set getSlots(Session session) throws Exception {
        Set slots = (Set) getCachedValue(_slotsCache);
        if (slots == null) {
            slots = getDelegate().getSlots(session);
            _slotsCache = createReference(slots);
        }
        return slots;
    }

    public Set getFacets(Session session) throws Exception {
        Set facets = (Set) getCachedValue(_facetsCache);
        if (facets == null) {
            facets = getDelegate().getFacets(session);
            _facetsCache = createReference(facets);
        }
        return facets;
    }

    public Set getFrames(Session session) throws Exception {
        return getDelegate().getFrames(session);
    }

    public Set getInstances(Cls cls, Session session) throws Exception {
        Set instances = (Set) getCachedValue(cls, _instancesCache);
        if (instances == null) {
            instances = getDelegate().getClses(session);
            setCachedValue(cls, instances, _instancesCache);
        }
        return instances;
    }

    public Set getReferences(Frame frame, Session session) throws Exception {
        Set references = (Set) getCachedValue(frame, _referencesCache);
        if (references == null) {
            references = getDelegate().getReferences(frame, session);
            setCachedValue(frame, references, _referencesCache);
        }
        return references;
    }

    public List getDirectOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        List values = (List) getCachedValue(frame, slot, _directOwnSlotValuesCache);
        if (values == null) {
            values = getDelegate().getDirectOwnSlotValues(frame, slot, session);
            setCachedValue(frame, slot, values, _directOwnSlotValuesCache);
        }
        return values;
    }

    public Collection getOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        Collection values = (List) getCachedValue(frame, slot, _ownSlotValuesCache);
        if (values == null) {
            values = getDelegate().getOwnSlotValues(frame, slot, session);
            setCachedValue(frame, slot, values, _ownSlotValuesCache);
        }
        return values;
    }

    public Frame getFrame(String name, Session session) throws Exception {
        Frame frame = (Frame) getCachedValue(name, _frameCache);
        if (frame == null) {
            frame = getDelegate().getFrame(name, session);
            setCachedValue(name, frame, _frameCache);
        }
        return frame;
    }

    public String getFrameName(Frame frame, Session session) throws Exception {
        String name = (String) getCachedValue(frame, _nameCache);
        if (name == null) {
            name = getDelegate().getFrameName(frame, session);
            setCachedValue(frame, name, _nameCache);
        }
        return name;
    }

    public Facet getAssociatedFacet(Slot slot, Session session) throws Exception {
        Facet facet = (Facet) getCachedValue(slot, _associatedFacetCache);
        if (facet == null) {
            facet = getDelegate().getAssociatedFacet(slot, session);
            setCachedValue(slot, facet, _associatedFacetCache);
        }
        return facet;
    }
    public Slot getAssociatedSlot(Facet facet, Session session) throws Exception {
        Slot slot = (Slot) getCachedValue(facet, _associatedSlotCache);
        if (slot == null) {
            slot = getDelegate().getAssociatedSlot(facet, session);
            setCachedValue(facet, slot, _associatedSlotCache);
        }
        return slot;
    }

    // modifications
    public void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values, Session session) throws Exception {
        getDelegate().setDirectOwnSlotValues(frame, slot, values, session);
    }

    public void deleteFrame(Frame frame, Session session) throws Exception {
        getDelegate().deleteFrame(frame, session);
    }

    public void setFrameName(Frame frame, String name, Session session) throws Exception {
        getDelegate().setFrameName(frame, name, session);
    }

    public Cls createCls(String name, Collection directTypes, Collection directSuperclasses, Session session) throws Exception {
        return getDelegate().createCls(name, directTypes, directSuperclasses, session);
    }

    public Slot createSlot(String name, Collection directTypes, Collection directSuperslots, Session session) throws Exception {
        return getDelegate().createSlot(name, directTypes, directSuperslots, session);
    }

    public Facet createFacet(String name, Collection directTypes, Session session) throws Exception {
        return getDelegate().createFacet(name, directTypes, session);
    }

    public SimpleInstance createSimpleInstance(String name, Collection directTypes, Session session) throws Exception {
        return getDelegate().createSimpleInstance(name, directTypes, session);
    }

    public void setAssociatedFacet(Slot slot, Facet facet, Session session) throws Exception {
        setAssociatedFacet(slot, facet, session);
    }

    public void setInverseSlot(Slot slotForward, Slot slotReverse, Session session) throws Exception {
        setInverseSlot(slotForward, slotReverse, session);
    }

    private void addEntries(Object value, Collection keys, Map map) {
        Iterator i = keys.iterator();
        while (i.hasNext()) {
            Object key = i.next();
            Collection c = (Collection) map.get(key);
            if (c != null) {
                c.add(value);
            }
        }
    }

    private void adjustForNewInstance(Instance instance, String name, Collection directTypes) {
        addEntries(instance, directTypes, _directInstancesCache);
        _instancesCache.clear();
        if (name != null) {
            setCachedValue(name, instance, _frameCache);
            setCachedValue(instance, name, _nameCache);
        }
    }

}

class Pair {
    private int _hashCode;
    private Object _o1;
    private Object _o2;

    public Pair(Object o1, Object o2) {
        set(o1, o2);
    }
    public int hashCode() {
        return _hashCode;
    }
    public boolean equals(Object o) {
        Pair pair = (Pair) o;
        return _o1.equals(pair._o1) && _o2.equals(pair._o2);
    }

    public void set(Object o1, Object o2) {
        _o1 = o1;
        _o2 = o2;
        _hashCode = HashUtils.getHash(o1, o2);
    }
}

class Triple {
    private int _hashCode;
    private Object _o1;
    private Object _o2;
    private Object _o3;

    public Triple(Object o1, Object o2, Object o3) {
        set(o1, o2, o3);
    }

    public int hashCode() {
        return _hashCode;
    }

    public boolean equals(Object o) {
        Triple triple = (Triple) o;
        return _o1.equals(triple._o1)
            && _o2.equals(triple._o2)
            && _o3.equals(triple._o3);
    }

    public void set(Object o1, Object o2, Object o3) {
        _o1 = o1;
        _o2 = o2;
        _o3 = o3;
        _hashCode = HashUtils.getHash(o1, o2, o3);
    }
}
*/
