/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.plugin;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;
import edu.stanford.smi.protege.widget.*;

/**
 * 
 * @author Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class PluginUtilities {
    private static Collection _manifestURLs = new HashSet();
    private static Collection _clsWidgetNames = new ArrayList();
    private static Collection _slotWidgetNames = new ArrayList();
    private static Collection _tabWidgetNames = new ArrayList();
    private static List _factoryNames = new ArrayList();
    private static Collection _importPluginNames = new ArrayList();
    private static Collection _exportPluginNames = new ArrayList();
    private static Collection _projectPluginNames = new ArrayList();
    private static Map _pluginClassNameToClassLoaderMap = new HashMap();
    private static Collection _factories;
    private static Map _defaultSlotWidgetNames = new HashMap();
    private static Set _pluginComponentNames = new HashSet();
    private static Map _pluginComponentNameToAboutURLMap = new HashMap();
    private static Map _pluginComponentNameToDocURLMap = new HashMap();
    private static Map _pluginPackageToClassLoaderMap = new HashMap();

    private static File pluginsDir;

    private static final PrintStream outputStream = System.out;

    private static final String PROPERTIES_FILE_NAME = "plugin.properties";
    private static final String PLUGIN_COUNT_PROPERTY = "plugin.component.count";
    private static final String PLUGIN_NAME_PROPERTY = "plugin.component.name";
    private static final String ABOUT_PROPERTY = "plugin.component.about";
    private static final String DOC_PROPERTY = "plugin.component.doc";
    private static final String DEPENDENCY_COUNT_PROPERTY = "plugin.dependency.count";
    private static final String DEPENDENCY_PROPERTY = "plugin.dependency";

    public final static String EXTRA_MANIFEST_PATH = "protege.plugin.manifest";

    private static FilenameFilter _pluginPackageFilter = new FilenameFilter() {
        public boolean accept(File dir, String name) {
            return new File(dir, name).isDirectory() && !name.equalsIgnoreCase("meta-inf") && !name.startsWith(".");
        }
    };

    static {
        init();
    }

    public static File getInstallationDirectory(String pluginClassName) {
        File directory = null;
        ClassLoader loader = (ClassLoader) _pluginClassNameToClassLoaderMap.get(pluginClassName);
        if (loader instanceof DirectoryClassLoader) {
            directory = ((DirectoryClassLoader) loader).getDirectory();
        }
        return directory;
    }

    public static void initialize() {
        // do nothing except ensure that the static initalizer has run
    }

    private static String attributeNameToClassName(String attributeName) {
        String className;
        if (attributeName.endsWith(".class")) {
            className = attributeName.substring(0, attributeName.length() - 6);
        } else {
            className = attributeName;
        }
        className = className.replace('/', '.');
        return className;
    }

    private static ClassLoader createClassLoader(File directory, ClassLoader parentLoader) {
        return new DirectoryClassLoader(directory, parentLoader);
    }

    public static Class forName(String className) {
        return forName(className, false);
    }

    /**
     * Load a class with the given name. The system first tries to load the class using the appropriate class loader
     * for a class with that name.  If this fails and promiscous=true then the system will try to load the class with
     * all available class loaders in no particular order until it finds one that works.
     */
    public static Class forName(String className, boolean promiscuous) {
        Class clas = null;
        try {
            ClassLoader loader = getClassLoader(className);
            Thread.currentThread().setContextClassLoader(loader);
            clas = Class.forName(className, true, loader);
        } catch (ClassNotFoundException e) {
            if (promiscuous) {
                clas = promiscuousForName(className);
            }
        } catch (Throwable e) {
            Log.exception(e, PluginUtilities.class, "forName", className);
        }
        return clas;
    }

    private static Class promiscuousForName(String className) {
        Class clas = null;
        // ClassLoader oldLoader = Thread.currentThread().getContextClassLoader();
        Iterator i = getClassLoaders().iterator();
        while (i.hasNext() && clas == null) {
            ClassLoader loader = (ClassLoader) i.next();
            Thread.currentThread().setContextClassLoader(loader);
            try {
                clas = Class.forName(className, true, loader);
            } catch (ClassNotFoundException e) {
                // do nothing
            } catch (NoClassDefFoundError error) {
                // Log.error(error.getMessage(), PluginUtilities.class, "promiscuousForName", className, loader);
            }
        }
        // Thread.currentThread().setContextClassLoader(oldLoader);
        return clas;
    }

    private static Collection getClassLoaders() {
        return new HashSet(_pluginClassNameToClassLoaderMap.values());
    }

    private static ClassLoader getClassLoader(String name) {
        ClassLoader loader = (ClassLoader) _pluginClassNameToClassLoaderMap.get(name);
        if (loader == null) {
            loader = PluginUtilities.class.getClassLoader();
        }
        return loader;
    }

    public static Collection getAvailableFactories() {
        if (_factories == null) {
            _factories = new ArrayList();
            Iterator i = _factoryNames.iterator();
            while (i.hasNext()) {
                String name = (String) i.next();
                _factories.add(SystemUtilities.newInstance(name));
            }
        }
        return _factories;
    }

    public static Collection getAvailableFactoryClassNames() {
        return Collections.unmodifiableCollection(_factoryNames);
    }

    public static Collection getAvailableSlotWidgetClassNames() {
        return Collections.unmodifiableCollection(_slotWidgetNames);
    }

    public static Collection getAvailableTabWidgetClassNames() {
        return Collections.unmodifiableCollection(_tabWidgetNames);
    }

    public static Collection getAvailableImportPluginClassNames() {
        return Collections.unmodifiableCollection(_importPluginNames);
    }

    public static Collection getAvailableExportPluginClassNames() {
        return Collections.unmodifiableCollection(_exportPluginNames);
    }

    public static Collection getAvailableProjectPluginClassNames() {
        return Collections.unmodifiableCollection(_projectPluginNames);
    }

    public static Collection getPluginComponentNames() {
        List list = new ArrayList(_pluginComponentNames);
        Collections.sort(list);
        return list;
    }

    public static URL getPluginComponentAboutURL(String pluginComponentName) {
        return (URL) _pluginComponentNameToAboutURLMap.get(pluginComponentName);
    }

    public static URL getPluginComponentDocURL(String pluginComponentName) {
        return (URL) _pluginComponentNameToDocURLMap.get(pluginComponentName);
    }

    public static String getDefaultWidgetClassName(boolean cardinality, ValueType type, Cls allowedCls) {
        DefaultEntry entry = new DefaultEntry(cardinality, type, allowedCls);
        String name = (String) _defaultSlotWidgetNames.get(entry);
        if (name == null && type.equals(ValueType.INSTANCE) && allowedCls != null) {
            entry.allowedClsName = null;
            name = (String) _defaultSlotWidgetNames.get(entry);
        }
        //Log.trace("returned name=" + name, SystemUtilities.class, "getDefaultWidgetClassName",
        //        new Boolean(cardinality), type, allowedCls);
        return name;
    }

    public static boolean isLoadableClass(String className) {
        return forName(className) != null;
    }

    private static boolean isLoadableClass(String className, ClassLoader loader, Class interfac) {
        boolean loadable = false;
        try {
            if (loader == null) {
                loader = SystemUtilities.class.getClassLoader();
            }
            Class clas = Class.forName(className, true, loader);
            if (clas != null) {
                loadable = interfac.isAssignableFrom(clas);
                if (!loadable) {
                    String text = className + " does not implement " + interfac;
                    Log.warning(text, PluginUtilities.class, "isLoadableClass");
                }
            }
        } catch (Throwable e) {
            Log.exception(e, PluginUtilities.class, "isLoadableClass");
        }
        return loadable;
    }

    private static boolean isSet(Attributes attributes, String name) {
        boolean isSet = false;
        String s = attributes.getValue(name);
        if (s != null) {
            isSet = s.equalsIgnoreCase("true");
        }
        return isSet;
    }

    private static void loadPluginsWithClassLoader(File file, ClassLoader classLoader) {
        printClassLoaderInformation(file, classLoader);
        Collection manifests = getNewManifests(classLoader);
        if (file == null) {
            Manifest extraManifest = loadExtraManifest();
            if (extraManifest != null) {
                manifests.add(extraManifest);
            }
        }
        Iterator i = manifests.iterator();
        while (i.hasNext()) {
            Manifest manifest = (Manifest) i.next();
            processManifest(manifest, classLoader);
        }
    }

    private static Collection getNewManifests(ClassLoader loader) {
        Collection manifests = new ArrayList();
        try {
            Enumeration e = loader.getResources("META-INF/MANIFEST.MF");
            while (e.hasMoreElements()) {
                URL url = (URL) e.nextElement();
                addNew(url, manifests);
            }
            // addNew(getExtraManifestURL(), manifests);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return manifests;
    }

    /*
    private static URL getExtraManifestURL() throws MalformedURLException {
        URL url = null;
        String path = System.getProperty(EXTRA_MANIFEST_PATH);
        if (path != null) {
            File file = new File(path);
            url = file.toURL();
        }
        return url;
    }
    */

    private static void addNew(URL manifestURL, Collection newManifests) throws IOException {
        if (manifestURL != null && _manifestURLs.add(manifestURL)) {
            newManifests.add(new Manifest(manifestURL.openStream()));
        }
    }

    private static void printClassLoaderInformation(File file, ClassLoader loader) {
        if (file != null && !file.equals(getPluginsDirectory())) {
            outputStream.println("    " + file.getName());
            if (false) {
                printURLs(loader);
            }
            outputStream.flush();
        }
    }

    private static void printURLs(ClassLoader loader) {
        URL[] urls = (loader instanceof URLClassLoader) ? ((URLClassLoader) loader).getURLs() : null;
        if (urls != null && urls.length != 0) {
            for (int i = 0; i < urls.length; ++i) {
                URL url = urls[i];
                outputStream.print("        ");
                outputStream.println(url);
            }
        }
    }

    public static void processManifest(Manifest manifest, ClassLoader loader) {
        Iterator i = manifest.getEntries().keySet().iterator();
        while (i.hasNext()) {
            String attributeName = (String) i.next();
            Attributes attributes = manifest.getAttributes(attributeName);
            String className = attributeNameToClassName(attributeName);
            checkPlugins(loader, attributes, className);
        }
    }

    private static void checkPlugins(ClassLoader loader, Attributes attributes, String className) {
        checkTabWidgetPlugin(loader, attributes, className);
        checkSlotWidgetPlugin(loader, attributes, className);
        checkClsWidgetPlugin(loader, attributes, className);
        checkFactoryPlugin(loader, attributes, className);
        checkImportPlugin(loader, attributes, className);
        checkExportPlugin(loader, attributes, className);
        checkProjectPlugin(loader, attributes, className);
    }

    private static void checkImportPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Import-Plugin")) {
            if (isLoadableClass(className, loader, ImportPlugin.class)) {
                addImportPlugin(className, loader);
            }
        }
    }

    private static void checkExportPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Export-Plugin")) {
            if (isLoadableClass(className, loader, ExportPlugin.class)) {
                addExportPlugin(className, loader);
            }
        }
    }

    private static void checkProjectPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Project-Plugin")) {
            if (isLoadableClass(className, loader, ProjectPlugin.class)) {
                addProjectPlugin(className, loader);
            }
        }
    }

    private static void checkFactoryPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Storage-Factory")) {
            if (isLoadableClass(className, loader, KnowledgeBaseFactory.class)) {
                boolean isDefault = isSet(attributes, "Default-Factory");
                addBackendPlugin(className, loader, isDefault);
            }
        }
    }

    private static void checkSlotWidgetPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Slot-Widget")) {
            if (isLoadableClass(className, loader, SlotWidget.class)) {
                addSlotWidgetPlugin(className, loader);
                if (isSet(attributes, "Default-Widget")) {
                    recordDefaults(className, attributes);
                }
            }
        }
    }

    private static void checkClsWidgetPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Cls-Widget")) {
            if (isLoadableClass(className, loader, ClsWidget.class)) {
                addClsWidgetPlugin(className, loader);
            }
        }
    }

    private static void checkTabWidgetPlugin(ClassLoader loader, Attributes attributes, String className) {
        if (isSet(attributes, "Tab-Widget")) {
            if (isLoadableClass(className, loader, TabWidget.class)) {
                addTabWidgetPlugin(className, loader);
            }
        }
    }

    public static void addExportPlugin(String className, ClassLoader loader) {
        _exportPluginNames.add(className);
        addPlugin(className, loader);
    }

    public static void addProjectPlugin(String className, ClassLoader loader) {
        _projectPluginNames.add(className);
        addPlugin(className, loader);
    }

    public static void addImportPlugin(String className, ClassLoader loader) {
        _importPluginNames.add(className);
        addPlugin(className, loader);
    }

    public static void addBackendPlugin(String className, ClassLoader loader, boolean isDefault) {
        if (isDefault) {
            _factoryNames.add(0, className);
        } else {
            _factoryNames.add(className);
        }
        addPlugin(className, loader);
    }

    public static void addClsWidgetPlugin(String className, ClassLoader loader) {
        _clsWidgetNames.add(className);
        addPlugin(className, loader);
    }

    public static void addSlotWidgetPlugin(String className, ClassLoader loader) {
        _slotWidgetNames.add(className);
        addPlugin(className, loader);
    }

    public static void addTabWidgetPlugin(String className, ClassLoader loader) {
        _tabWidgetNames.add(className);
        addPlugin(className, loader);
    }

    private static void addPlugin(String pluginClassName, ClassLoader loader) {
        Object o = _pluginClassNameToClassLoaderMap.put(pluginClassName, loader);
        if (o != null && o != loader) {
            Log.error("Duplicate", PluginUtilities.class, "addPlugin", pluginClassName);
        }
    }

    private static void recordDefaults(String className, Attributes attributes) {
        String cardinality = attributes.getValue("Default-Widget-For-Cardinality");
        String type = attributes.getValue("Default-Widget-For-Type");
        String allowed_class = attributes.getValue("Default-Widget-For-Allowed-Class");
        DefaultEntry entry = new DefaultEntry(cardinality, type, allowed_class);
        _defaultSlotWidgetNames.put(entry, className);
    }

    public static File getPluginsDirectory() {
        File pluginsDir = null;
        File dir = ApplicationProperties.getApplicationDirectory();
        if (dir == null) {
            Log.warning("Application directory not specified", PluginUtilities.class, "getPluginsDir");
        } else {
            File file = new File(dir, "plugins");
            if (file.exists()) {
                pluginsDir = file;
            } else {
                Log.warning("Plugins directory not found: " + file, PluginUtilities.class, "getPluginsDir");
            }
        }
        return pluginsDir;
    }

    private static void init() {
        try {
            loadPlugins();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static void loadPlugins() {
        loadSystemPlugins();

        pluginsDir = getPluginsDirectory();
        if (pluginsDir != null) {
            System.out.println("installed plugins:");
            loadLegacyPlugins(pluginsDir);
            loadPluginPackages(pluginsDir);
        }
        // loadExtraManifestPlugins(pluginsDir);
    }

    private static void loadSystemPlugins() {
        loadPlugins(null);
    }

    private static void loadPlugins(File dir) {
        createClassLoaderAndLoadPlugins(dir, PluginUtilities.class.getClassLoader());
    }

    private static void createClassLoaderAndLoadPlugins(File dir, ClassLoader parentLoader) {
        ClassLoader loader = createClassLoader(dir, parentLoader);
        _pluginPackageToClassLoaderMap.put(dir, loader);
        loadPluginsWithClassLoader(dir, loader);
    }

    private static void loadLegacyPlugins(File pluginsDir) {
        loadPlugins(pluginsDir);
    }

    private static ClassLoader getParentLoader(File packageDir) {
        ClassLoader parentLoader;
        Collection dependentLoaders = getDependentLoaders(packageDir);
        int count = dependentLoaders.size();
        switch (count) {
            case 0 :
                parentLoader = ClassLoader.getSystemClassLoader();
                break;
            case 1 :
                parentLoader = (ClassLoader) CollectionUtilities.getFirstItem(dependentLoaders);
                break;
            default :
                parentLoader = new MultiplexingClassLoader(dependentLoaders);
                break;
        }
        return parentLoader;
    }

    private static Collection getDependentLoaders(File packageDir) {
        Collection dependentLoaders = new ArrayList();
        Properties properties = getProperties(packageDir);
        int count = getInt(properties, DEPENDENCY_COUNT_PROPERTY);
        for (int i = 0; i < count; ++i) {
            File dependentPackage = getDependentPackage(properties, i);
            if (dependentPackage != null) {
                ClassLoader packageLoader = getClassLoaderForPackage(dependentPackage);
                if (packageLoader != null) {
                    dependentLoaders.add(packageLoader);
                }
            }
        }
        return dependentLoaders;
    }

    private static ClassLoader getClassLoaderForPackage(File file) {
        ClassLoader loader = (ClassLoader) _pluginPackageToClassLoaderMap.get(file);
        if (loader == null) {
            loadPluginPackage(file);
            loader = (ClassLoader) _pluginPackageToClassLoaderMap.get(file);
        }
        return loader;
    }

    private static File getDependentPackage(Properties properties, int i) {
        File dependentPackage = null;
        String pack = properties.getProperty(DEPENDENCY_PROPERTY + "." + i);
        if (pack != null) {
            File testPackage = new File(pluginsDir, pack);
            if (testPackage.isDirectory()) {
                dependentPackage = testPackage;
            }
        }
        return dependentPackage;
    }

    private static void loadPluginPackages(File pluginsDir) {
        File[] packages = pluginsDir.listFiles(_pluginPackageFilter);
        Collection list = orderPackages(packages);
        Iterator i = list.iterator();
        while (i.hasNext()) {
            File packageDir = (File) i.next();
            if (!isLoaded(packageDir)) {
                loadPluginPackage(packageDir);
            }
        }
    }

    private static boolean isLoaded(File packageDir) {
        return _pluginPackageToClassLoaderMap.get(packageDir) != null;
    }

    private static Collection orderPackages(File[] packages) {
        return Arrays.asList(packages);
    }

    private static void loadPluginPackage(File packageDir) {
        createClassLoaderAndLoadPlugins(packageDir, getParentLoader(packageDir));
        loadURLs(packageDir);
    }

    private static Manifest loadExtraManifest() {
        Manifest manifest = null;
        String fileName = ApplicationProperties.getExtraManifestPath();
        if (fileName != null) {
            try {
                InputStream is = new FileInputStream(fileName);
                manifest = new Manifest(is);
            } catch (IOException e) {
                Log.exception(e, PluginUtilities.class, "loadExtraManifest");
            }
        }
        return manifest;
    }

    private static int getInt(Properties properties, String property) {
        int count = 0;
        if (properties != null) {
            String countString = properties.getProperty(property);
            if (countString != null) {
                count = Integer.parseInt(countString);
            }
        }
        return count;
    }

    private static String getPluginName(Properties properties, int i, File dir) {
        String pluginName = PLUGIN_NAME_PROPERTY + "." + i;
        return properties.getProperty(pluginName, dir.getName());
    }

    private static void loadURL(String name, Properties properties, String property, int i, File dir, Map map) {
        String filePropertyName = property + "." + i;
        String fileString = properties.getProperty(filePropertyName);
        if (fileString != null) {
            URL url = getURL(dir, fileString);
            if (url == null) {
                Log.error("missing file: " + fileString, PluginUtilities.class, "loadURL", name, property);
            } else {
                map.put(name, url);
                // Log.trace("loaded: " + url, PluginUtilities.class, "loadURL", name, property);
            }
        }
    }

    private static final char STANDARD_SEPARATOR_CHAR = '/';
    private static String localizePath(String path) {
        char replacement = File.separatorChar;
        if (replacement != STANDARD_SEPARATOR_CHAR) {
            path = path.replace(STANDARD_SEPARATOR_CHAR, replacement);
        }
        return path;
    }

    private static URL getURL(File dir, String path) {
        URL url = null;
        URI uri = getAbsoluteURI(path);
        if (uri == null) {
            uri = getAbsoluteURI(dir, path);
        }
        if (uri != null) {
            try {
                url = uri.toURL();
            } catch (MalformedURLException e) {
            }
        }
        return url;
    }

    private static URI getAbsoluteURI(File dir, String path) {
        URI uri = null;
        path = localizePath(path);
        File file = new File(dir, path);
        if (file.exists()) {
            uri = file.toURI();
        }
        return uri;
    }

    private static URI getAbsoluteURI(String path) {
        URI uri = null;
        try {
            uri = new URI(path);
            if (!uri.isAbsolute()) {
                uri = null;
            }
        } catch (URISyntaxException e) {
            // do nothing
        }
        return uri;
    }

    private static Properties getProperties(File dir) {
        Properties properties = null;
        File file = new File(dir, PROPERTIES_FILE_NAME);
        if (file.exists()) {
            try {
                InputStream inputStream = new FileInputStream(file);
                properties = new Properties();
                properties.load(inputStream);
                inputStream.close();
            } catch (IOException e) {
                Log.exception(e, PluginUtilities.class, "getProperties", dir);
            }
        }
        return properties;
    }

    private static void loadURLs(File dir) {
        Properties properties = getProperties(dir);
        if (properties != null) {
            int count = getInt(properties, PLUGIN_COUNT_PROPERTY);
            for (int i = 0; i < count; ++i) {
                String pluginName = getPluginName(properties, i, dir);
                _pluginComponentNames.add(pluginName);
                loadURL(pluginName, properties, ABOUT_PROPERTY, i, dir, _pluginComponentNameToAboutURLMap);
                loadURL(pluginName, properties, DOC_PROPERTY, i, dir, _pluginComponentNameToDocURLMap);
            }
        }
    }

    /**
     *  Does a search of the available manifests entries for the specified
     *  attribute key and does a case insensitive match on the specified
     *  attribute value. If there is a match the associated java Class object is
     *  loaded.
     */
    public static Collection getClassesWithAttribute(String key, String value) {
        // Log.enter(SystemUtilities.class, "getManifestClasses", key, value);
        Collection classes = new HashSet();
        Iterator i = _manifestURLs.iterator();
        while (i.hasNext()) {
            URL url = (URL) i.next();
            try {
                Manifest manifest = new Manifest(url.openStream());
                classes.addAll(getManifestClasses(manifest, key, value));
            } catch (IOException e) {
                Log.exception(e, PluginUtilities.class, "getClassesWithAttribute", key, value);
            }
        }
        return classes;
    }

    private static Collection getManifestClasses(Manifest manifest, String key, String value) {
        Collection classes = new HashSet();
        Iterator i = manifest.getEntries().keySet().iterator();
        while (i.hasNext()) {
            String attributeName = (String) i.next();
            Attributes attributes = manifest.getAttributes(attributeName);
            String attributeValue = attributes.getValue(key);
            if (equalsAttributes(attributeValue, value)) {
                String className = attributeNameToClassName(attributeName);
                Class clas = forName(className, true);
                if (clas != null) {
                    classes.add(clas);
                }
            }
        }
        return classes;
    }

    private static boolean equalsAttributes(String value1, String value2) {
        return (value1 == null) ? (value2 == null) : value1.equalsIgnoreCase(value2);
    }
}
