/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.server;

import java.rmi.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.model.query.*;

public interface RemoteServerFrameStore extends Remote {

    int getClsCount(RemoteSession session) throws RemoteException;
    int getSlotCount(RemoteSession session) throws RemoteException;
    int getFacetCount(RemoteSession session) throws RemoteException;
    int getSimpleInstanceCount(RemoteSession session) throws RemoteException;
    int getFrameCount(RemoteSession session) throws RemoteException;

    // frame access
    Set getClses(RemoteSession session) throws RemoteException;
    Set getSlots(RemoteSession session) throws RemoteException;
    Set getFacets(RemoteSession session) throws RemoteException;
    Set getFrames(RemoteSession session) throws RemoteException;
    Frame getFrame(String name, RemoteSession session) throws RemoteException;
    String getFrameName(Frame frame, RemoteSession session) throws RemoteException;
    void setFrameName(Frame frame, String name, RemoteSession session) throws RemoteException;

    // frame creation/deletion
    Cls createCls(
        FrameID id,
        String name,
        Collection directTypes,
        Collection directSuperclasses,
        boolean loadDefaultValues,
        RemoteSession session)
        throws RemoteException;
    Slot createSlot(
        FrameID id,
        String name,
        Collection directTypes,
        Collection directSuperslots,
        boolean loadDefaultValues,
        RemoteSession session)
        throws RemoteException;
    Facet createFacet(
        FrameID id,
        String name,
        Collection directTypes,
        boolean loadDefaultValues,
        RemoteSession session)
        throws RemoteException;
    SimpleInstance createSimpleInstance(
        FrameID id,
        String name,
        Collection directTypes,
        boolean loadDefaultValues,
        RemoteSession session)
        throws RemoteException;

    void deleteCls(Cls cls, RemoteSession session) throws RemoteException;
    void deleteSlot(Slot slot, RemoteSession session) throws RemoteException;
    void deleteFacet(Facet facet, RemoteSession session) throws RemoteException;
    void deleteSimpleInstance(SimpleInstance simpleInstance, RemoteSession session) throws RemoteException;

    // own slots
    Set getOwnSlots(Frame frame, RemoteSession session) throws RemoteException;
    Collection getOwnSlotValues(Frame frame, Slot slot, RemoteSession session) throws RemoteException;
    List getDirectOwnSlotValues(Frame frame, Slot slot, RemoteSession session) throws RemoteException;
    int getDirectOwnSlotValuesCount(Frame frame, Slot slot, RemoteSession session) throws RemoteException;
    void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values, RemoteSession session)
        throws RemoteException;

    // own facets
    Set getOwnFacets(Frame frame, Slot slot, RemoteSession session) throws RemoteException;
    Collection getOwnFacetValues(Frame frame, Slot slot, Facet facet, RemoteSession session) throws RemoteException;

    // template slots
    Set getTemplateSlots(Cls cls, RemoteSession session) throws RemoteException;
    List getDirectTemplateSlots(Cls cls, RemoteSession session) throws RemoteException;
    List getDirectDomain(Slot slot, RemoteSession session) throws RemoteException;
    Set getDomain(Slot slot, RemoteSession session) throws RemoteException;
    // Set getInheritedTemplateSlots(Cls cls, Session session) throws RemoteException;
    Set getOverriddenTemplateSlots(Cls cls, RemoteSession session) throws RemoteException;
    Set getDirectlyOverriddenTemplateSlots(Cls cls, RemoteSession session) throws RemoteException;
    void addDirectTemplateSlot(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    void removeDirectTemplateSlot(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    void moveDirectTemplateSlot(Cls cls, Slot slot, int index, RemoteSession session) throws RemoteException;

    // template slot values
    Collection getTemplateSlotValues(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    List getDirectTemplateSlotValues(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    void setDirectTemplateSlotValues(Cls cls, Slot slot, Collection values, RemoteSession session)
        throws RemoteException;

    // template facets
    Set getTemplateFacets(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    Set getOverriddenTemplateFacets(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    Set getDirectlyOverriddenTemplateFacets(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    void removeDirectTemplateFacetOverrides(Cls cls, Slot slot, RemoteSession session) throws RemoteException;
    Collection getTemplateFacetValues(Cls cls, Slot slot, Facet facet, RemoteSession session) throws RemoteException;
    List getDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, RemoteSession session) throws RemoteException;
    void setDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Collection values, RemoteSession session)
        throws RemoteException;

    // class hierarchy
    List getDirectSuperclasses(Cls cls, RemoteSession session) throws RemoteException;
    Set getSuperclasses(Cls cls, RemoteSession session) throws RemoteException;
    List getDirectSubclasses(Cls cls, RemoteSession session) throws RemoteException;
    Set getSubclasses(Cls cls, RemoteSession session) throws RemoteException;
    void addDirectSuperclass(Cls cls, Cls superclass, RemoteSession session) throws RemoteException;
    void removeDirectSuperclass(Cls cls, Cls superclass, RemoteSession session) throws RemoteException;
    void moveDirectSubclass(Cls cls, Cls subclass, int index, RemoteSession session) throws RemoteException;

    // slot hierarchy
    List getDirectSuperslots(Slot slot, RemoteSession session) throws RemoteException;
    Set getSuperslots(Slot slot, RemoteSession session) throws RemoteException;
    List getDirectSubslots(Slot slot, RemoteSession session) throws RemoteException;
    Set getSubslots(Slot slot, RemoteSession session) throws RemoteException;
    void addDirectSuperslot(Slot slot, Slot superslot, RemoteSession session) throws RemoteException;
    void removeDirectSuperslot(Slot slot, Slot superslot, RemoteSession session) throws RemoteException;
    void moveDirectSubslot(Slot slot, Slot subslot, int index, RemoteSession session) throws RemoteException;

    // type hierarchy
    List getDirectTypes(Instance instance, RemoteSession session) throws RemoteException;
    Set getTypes(Instance instance, RemoteSession session) throws RemoteException;
    List getDirectInstances(Cls cls, RemoteSession session) throws RemoteException;
    Set getInstances(Cls cls, RemoteSession session) throws RemoteException;
    void addDirectType(Instance instance, Cls type, RemoteSession session) throws RemoteException;
    void removeDirectType(Instance instance, Cls type, RemoteSession session) throws RemoteException;

    // events
    List getEvents(RemoteSession session) throws RemoteException;

    // arbitrary queries
    Set getReferences(Object object, RemoteSession session) throws RemoteException;
    Set getMatchingReferences(String string, int maxMatches, RemoteSession session) throws RemoteException;
    Set getFramesWithDirectOwnSlotValue(Slot slot, Object value, RemoteSession session) throws RemoteException;
    Set getFramesWithAnyDirectOwnSlotValue(Slot slot, RemoteSession session) throws RemoteException;
    Set getFramesWithMatchingDirectOwnSlotValue(Slot slot, String value, int maxMatches, RemoteSession session)
        throws RemoteException;
    Set getClsesWithDirectTemplateSlotValue(Slot slot, Object value, RemoteSession session) throws RemoteException;
    Set getClsesWithAnyDirectTemplateSlotValue(Slot slot, RemoteSession session) throws RemoteException;
    Set getClsesWithMatchingDirectTemplateSlotValue(Slot slot, String value, int maxMatches, RemoteSession session)
        throws RemoteException;
    Set getClsesWithDirectTemplateFacetValue(Slot slot, Facet facet, Object value, RemoteSession session)
        throws RemoteException;
    Set getClsesWithMatchingDirectTemplateFacetValue(
        Slot slot,
        Facet facet,
        String value,
        int maxMatches,
        RemoteSession session)
        throws RemoteException;
    Set executeQuery(Query query, RemoteSession session) throws RemoteException;

    // closures
    Set getDirectOwnSlotValuesClosure(Frame frame, Slot slot, RemoteSession session) throws RemoteException;

    void close(RemoteSession session) throws RemoteException;
    boolean beginTransaction(String name, RemoteSession session) throws RemoteException;
    boolean commitTransaction(RemoteSession session) throws RemoteException;
    boolean rollbackTransaction(RemoteSession session) throws RemoteException;
    Frame getFrame(FrameID id, RemoteSession session) throws RemoteException;
}
