/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.ui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.resource.*;
import edu.stanford.smi.protege.util.*;
import edu.stanford.smi.protege.widget.*;

/**
 * The GUI view of a project.  This is basically the outer tabbed pane for the application.
 *
 * @author    Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class ProjectView extends JComponent {
    private Project _project;
    private JTabbedPane _tabbedPane;
    private Collection _currentClsPath;
    private boolean _autosynchronizeTrees;
    private Collection _detachedTabs = new HashSet();

    public ProjectView(Project project) {
        _project = project;
        setLayout(new BorderLayout());
        add(createTabbedPane(), BorderLayout.CENTER);
        project.getKnowledgeBase().setUndoEnabled(true);
    }

    private void addTab(WidgetDescriptor widgetDescriptor) {
        TabWidget widget = WidgetUtilities.createTabWidget(widgetDescriptor, _project);
        addTab(widget);
    }

    private void addTab(TabWidget widget) {
        addTab(widget, -1);
    }
    private void addTab(final TabWidget widget, int index) {
        JComponent component = (JComponent) widget;
        Icon icon = widget.getIcon();
        String title = widget.getLabel();
        String help = widget.getShortDescription();
        if (index < 0) {
            _tabbedPane.addTab(title, icon, component, help);
        } else {
            _tabbedPane.insertTab(title, icon, component, help, index);
        }
        widget.addSelectionListener(new SelectionListener() {
            public void selectionChanged(SelectionEvent event) {
                java.util.List list = new ArrayList(widget.getSelection());
                if (!list.isEmpty() && list.get(0) instanceof Collection) {
                    list.remove(0);
                }
                setCurrentClsPath(list);
            }
        });
    }

    public boolean attemptClose() {
        boolean canClose = true;
        Iterator i = getTabs().iterator();
        while (canClose && i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            canClose = tab.canClose();
        }
        if (canClose) {
            close();
        }
        return canClose;
    }

    public boolean attemptSave() {
        boolean canSave = true;
        Iterator i = getTabs().iterator();
        while (canSave && i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            canSave = tab.canSave();
        }
        if (canSave) {
            save();
        }
        return canSave;
    }

    public void close() {
        Iterator i = getTabs().iterator();
        while (i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            tab.close();
        }
    }

    private JComponent createTabbedPane() {
        _tabbedPane = ComponentFactory.createTabbedPane(true);
        Iterator i = _project.getTabWidgetDescriptors().iterator();
        while (i.hasNext()) {
            WidgetDescriptor d = (WidgetDescriptor) i.next();
            if (d.isVisible()) {
                addTab(d);
            }
        }
        if (_tabbedPane.getTabCount() > 0) {
            _tabbedPane.setSelectedIndex(0);
        }
        return _tabbedPane;
    }

    public void dispose() {
    }

    public Project getProject() {
        return _project;
    }

    /**
     * Gets a tab by its full class name (not it label!)
     */
    public TabWidget getTab(String className) {
        TabWidget result = null;
        Component[] components = _tabbedPane.getComponents();
        for (int i = 0; i < components.length; ++i) {
            Component c = components[i];
            if (c.getClass().getName().equals(className)) {
                result = (TabWidget) c;
                break;
            }
        }
        return result;
    }

    private Collection getTabs() {
        return Arrays.asList(_tabbedPane.getComponents());
    }

    public void reload(boolean regenerate) {
        if (regenerate) {
            reloadAll();
        } else {
            reloadPartial();
        }
    }

    private void reloadPartial() {
        closeDetachedDisabledTabs();
        synchronizeTabbedPane();
    }

    private void synchronizeTabbedPane() {
        removeDisabledTabs();
        addEnabledTabs();
        reorderTabs();
    }

    private void removeDisabledTabs() {
        Iterator i = getTabs().iterator();
        while (i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            if (!isEnabled(tab)) {
                _tabbedPane.remove((Component) tab);
                tab.dispose();
            }
        }
    }

    private void addEnabledTabs() {
        Iterator i = _project.getTabWidgetDescriptors().iterator();
        while (i.hasNext()) {
            WidgetDescriptor d = (WidgetDescriptor) i.next();
            if (d.isVisible()) {
                String className = d.getWidgetClassName();
                if (getTab(className) == null) {
                    addTab(d);
                }
            }
        }
    }

    private void reorderTabs() {
        int index = 0;
        Iterator i = _project.getTabWidgetDescriptors().iterator();
        while (i.hasNext()) {
            WidgetDescriptor d = (WidgetDescriptor) i.next();
            if (d.isVisible()) {
                int currentIndex = getTabIndex(d);
                if (currentIndex != index) {
                    TabWidget tab = (TabWidget) _tabbedPane.getComponent(currentIndex);
                    addTab(tab, index);
                }
                ++index;
            }
        }
    }

    private int getTabIndex(WidgetDescriptor d) {
        return _tabbedPane.indexOfComponent((Component) getTab(d.getWidgetClassName()));
    }

    public void reloadAll() {
        closeDetachedTabs();
        if (_tabbedPane != null) {
            ComponentUtilities.dispose(_tabbedPane);
        }
        removeAll();
        _project.clearCachedWidgets();
        add(createTabbedPane());
        revalidate();
        repaint();
    }

    private void closeDetachedTabs() {
        Iterator i = _detachedTabs.iterator();
        while (i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            i.remove();
            closeDetachedTab(tab);
        }
    }

    private void closeDetachedTab(TabWidget tab) {
        Component c = (Component) tab;
        JFrame frame = (JFrame) SwingUtilities.getRoot(c);
        frame.dispose();
    }

    private void closeDetachedDisabledTabs() {
        Iterator i = _detachedTabs.iterator();
        while (i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            if (!isEnabled(tab)) {
                i.remove();
                closeDetachedTab(tab);
            }
        }
    }

    private boolean isEnabled(TabWidget tab) {
        return tab.getDescriptor().isVisible();
    }

    private void save() {
        Iterator i = getTabs().iterator();
        while (i.hasNext()) {
            TabWidget tab = (TabWidget) i.next();
            tab.save();
        }
    }

    public String toString() {
        return "ProjectView";
    }

    public void detachCurrentTab() {
        JComponent c = (JComponent) _tabbedPane.getSelectedComponent();
        int index = _tabbedPane.getSelectedIndex();
        _tabbedPane.removeTabAt(index);
        final JFrame frame = ComponentFactory.createFrame();
        c.setPreferredSize(c.getSize());
        frame.getContentPane().add(c);
        frame.pack();
        String title = ((TabWidget) c).getLabel() + " Tab - " + Text.getProgramName();
        frame.setTitle(title);
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                Component c = frame.getContentPane().getComponent(0);
                reattachTab(c);
                frame.setVisible(false);
                frame.dispose();
            }
        });
        frame.setVisible(true);
        _detachedTabs.add(c);
    }

    private void reattachTab(Component c) {
        int index = getInsertionPoint(c);
        addTab((TabWidget) c, index);
        _tabbedPane.setSelectedIndex(index);
        _detachedTabs.remove(c);
    }

    private int getInsertionPoint(Component c) {
        String classNameToMatch = c.getClass().getName();
        int insertionPoint = 0;
        Iterator i = _project.getTabWidgetDescriptors().iterator();
        while (i.hasNext() && insertionPoint < _tabbedPane.getComponentCount()) {
            WidgetDescriptor desc = (WidgetDescriptor) i.next();
            String className = desc.getWidgetClassName();
            if (className.equals(classNameToMatch)) {
                break;
            }
            if (className.equals(_tabbedPane.getComponentAt(insertionPoint).getClass().getName())) {
                ++insertionPoint;
            }
        }
        return insertionPoint;
    }

    public void setAutosynchronizeClsTrees(boolean b) {
        _autosynchronizeTrees = b;
        synchronizeClsTree();
    }

    private void setCurrentClsPath(Collection c) {
        _currentClsPath = c;
        if (_autosynchronizeTrees) {
            Iterator i = Arrays.asList(_tabbedPane.getComponents()).iterator();
            while (i.hasNext()) {
                TabWidget tab = (TabWidget) i.next();
                tab.synchronizeClsTree(_currentClsPath);
            }
        }
    }

    public void synchronizeClsTree() {
        if (_currentClsPath != null) {
            TabWidget currentWidget = (TabWidget) _tabbedPane.getSelectedComponent();
            if (currentWidget == null) {
                Log.trace("strange, no selected tab", this, "synchronizeClsTree");
            } else {
                currentWidget.synchronizeClsTree(_currentClsPath);
            }
        }
    }
}
